from baby_steps import given, then, when
from niltype import Nil

from district42 import Props


def test_props_set():
    with given:
        props = Props()

    with when:
        res = props.set("key", "val")

    with then:
        assert isinstance(res, Props)
        assert res != props


def test_props_update():
    with given:
        props = Props()

    with when:
        res = props.update(key1="val1", key2="val2")

    with then:
        assert isinstance(res, Props)
        assert res != props


def test_props_get_default():
    with given:
        props = Props()

    with when:
        res = props.get("non_existing")

    with then:
        assert res is Nil


def test_props_get_custom_default():
    with given:
        props = Props()

    with when:
        res = props.get("non_existing", None)

    with then:
        assert res is None


def test_get_setted():
    with given:
        key, val = "key", "val"
        props = Props().set(key, val)

    with when:
        res = props.get("key")

    with then:
        assert res == val


def test_get_updated():
    with given:
        props = Props().update(key1="val1", key2="val2")

    with when:
        res = props.get("key1")

    with then:
        assert res == "val1"


def test_props_repr():
    with given:
        props = Props()

    with when:
        res = repr(props)

    with then:
        assert res == "<Props {}>"


def test_props_eq_empty():
    with given:
        props1 = Props()
        props2 = Props()

    with when:
        res = props1 == props2

    with then:
        assert res is True


def test_props_eq_updated():
    with given:
        props1 = Props().update(key1="val1", key2="val2")
        props2 = Props().update(key1="val1", key2="val2")

    with when:
        res = props1 == props2

    with then:
        assert res is True


def test_props_not_eq_incorrect_other():
    with given:
        props1 = Props().update(key1="val1", key2="val2")

    with when:
        res = props1 == "Props()"

    with then:
        assert res is False


def test_props_not_eq_less_other():
    with given:
        props1 = Props().update(key1="val1", key2="val2")
        props2 = Props().update(key1="val1")

    with when:
        res = props1 == props2

    with then:
        assert res is False


def test_props_not_eq_more_other():
    with given:
        props1 = Props().update(key1="val1")
        props2 = Props().update(key1="val1", key2="val2")

    with when:
        res = props1 == props2

    with then:
        assert res is False


def test_props_no_keys():
    with given:
        props = Props()

    with when:
        keys = [key for key in props]

    with then:
        assert keys == []


def test_props_keys():
    with given:
        props = Props().update(key1="val1", key2="val2")

    with when:
        keys = [key for key in props]

    with then:
        assert keys == ["key1", "key2"]



from baby_steps import then, when

from district42 import schema, union


def test_union():
    with when:
        res = union(schema.str, schema.none)

    with then:
        assert res == schema.any(schema.str, schema.none)


def test_union_nested():
    with when:
        res = union(union(schema.str, schema.int), schema.none)

    with then:
        assert res == schema.any(schema.any(schema.str, schema.int), schema.none)



from typing import Any

from niltype import Nil

from .._schema_visitor import SchemaVisitor
from ..types import (
    AnySchema,
    BoolSchema,
    BytesSchema,
    ConstSchema,
    DateTimeSchema,
    DictSchema,
    FloatSchema,
    GenericTypeAliasSchema,
    IntSchema,
    ListSchema,
    NoneSchema,
    StrSchema,
    TypeAliasPropsType,
    UUID4Schema,
)
from ..utils import is_ellipsis


class Representor(SchemaVisitor[str]):
    def __init__(self, name: str = "schema", indent: int = 4) -> None:
        self._name = name
        self._indent = indent

    def visit_none(self, schema: NoneSchema, *, indent: int = 0, **kwargs: Any) -> str:
        return f"{self._name}.none"

    def visit_bool(self, schema: BoolSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.bool"

        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"
        return r

    def visit_int(self, schema: IntSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.int"

        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"

        if schema.props.min is not Nil:
            r += f".min({schema.props.min!r})"

        if schema.props.max is not Nil:
            r += f".max({schema.props.max!r})"

        return r

    def visit_float(self, schema: FloatSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.float"

        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"

        if schema.props.min is not Nil:
            r += f".min({schema.props.min!r})"

        if schema.props.max is not Nil:
            r += f".max({schema.props.max!r})"

        if schema.props.precision is not Nil:
            r += f".precision({schema.props.precision!r})"

        return r

    def visit_str(self, schema: StrSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.str"

        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"

        if schema.props.alphabet is not Nil:
            r += f".alphabet({schema.props.alphabet!r})"

        if schema.props.substr is not Nil:
            r += f".contains({schema.props.substr!r})"

        if schema.props.pattern is not Nil:
            r += f".regex({schema.props.pattern!r})"

        if schema.props.len is not Nil:
            r += f".len({schema.props.len!r})"
        elif (schema.props.min_len is not Nil) and (schema.props.max_len is not Nil):
            r += f".len({schema.props.min_len!r}, {schema.props.max_len!r})"
        elif schema.props.min_len is not Nil:
            r += f".len({schema.props.min_len!r}, ...)"
        elif schema.props.max_len is not Nil:
            r += f".len(..., {schema.props.max_len!r})"

        return r

    def visit_list(self, schema: ListSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.list"

        if schema.props.type is not Nil:
            r += "({})".format(schema.props.type.__accept__(self, indent=indent, **kwargs))
        elif schema.props.elements is not Nil:
            if len(schema.props.elements) == 0:
                return r + "([])"
            elems = []
            for element in schema.props.elements:
                if is_ellipsis(element):
                    elem = "..."
                else:
                    elem = element.__accept__(self, indent=indent + self._indent, **kwargs)
                elems.append(" " * (indent + self._indent) + elem)
            r += "([\n"
            r += ",\n".join(elems)
            r += "\n" + " " * indent + "])"

        if schema.props.len is not Nil:
            r += f".len({schema.props.len!r})"
        elif (schema.props.min_len is not Nil) and (schema.props.max_len is not Nil):
            r += f".len({schema.props.min_len!r}, {schema.props.max_len!r})"
        elif schema.props.min_len is not Nil:
            r += f".len({schema.props.min_len!r}, ...)"
        elif schema.props.max_len is not Nil:
            r += f".len(..., {schema.props.max_len!r})"

        return r

    def visit_dict(self, schema: DictSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.dict"

        if schema.props.keys is Nil:
            return r

        if len(schema.props.keys) == 0:
            return r + "({})"
        elif (len(schema.props.keys) == 1) and (... in schema.props.keys):
            return r + "({...: ...})"

        pairs = []
        for key, (val, is_optional) in schema.props.keys.items():
            if is_ellipsis(key):
                key_repr = val_repr = "..."
            else:
                key_repr = f"optional({key!r})" if is_optional else repr(key)
                val_repr = val.__accept__(self, indent=indent + self._indent, **kwargs)
            pairs.append("{indent}{key}: {val}".format(
                indent=" " * (indent + self._indent),
                key=key_repr,
                val=val_repr,
            ))

        r += "({\n"
        r += ",\n".join(pairs) + "\n"
        r += " " * indent + "})"

        return r

    def visit_any(self, schema: AnySchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.any"

        if schema.props.types is not Nil:
            types = [x.__accept__(self, indent=indent, **kwargs) for x in schema.props.types]
            r += "({})".format(", ".join(types))
        return r

    def visit_const(self, schema: ConstSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.const"

        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"
        return r

    def visit_bytes(self, schema: BytesSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.bytes"

        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"
        return r

    def visit_type_alias(self, schema: GenericTypeAliasSchema[TypeAliasPropsType],
                         *, indent: int = 0, **kwargs: Any) -> str:
        type_name = schema.__class__.__name__
        type_repr = schema.props.type.__accept__(self, indent=indent, **kwargs)
        if schema.props.name is not Nil:
            type_name = schema.props.name
        return f"{type_name}<{type_repr}>"

    def visit_uuid4(self, schema: UUID4Schema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.uuid4"

        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"

        return r

    def visit_datetime(self, schema: DateTimeSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.datetime"

        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"

        return r



import warnings
from typing import Any

from .types import (
    AnySchema,
    BoolSchema,
    BytesSchema,
    ConstSchema,
    DateTimeSchema,
    DictSchema,
    FloatSchema,
    GenericSchema,
    IntSchema,
    ListSchema,
    NoneSchema,
    StrSchema,
    TypeAliasProps,
    TypeAliasSchema,
    UUID4Schema,
)

__all__ = ("SchemaFacade",)


class SchemaFacade:
    def alias(self, /, name: str, type_: GenericSchema) -> TypeAliasSchema:
        props = TypeAliasProps()
        return TypeAliasSchema(props.update(name=name, type=type_))

    @property
    def none(self) -> NoneSchema:
        return NoneSchema()

    @property
    def bool(self) -> BoolSchema:
        return BoolSchema()

    @property
    def int(self) -> IntSchema:
        return IntSchema()

    @property
    def float(self) -> FloatSchema:
        return FloatSchema()

    @property
    def str(self) -> StrSchema:
        return StrSchema()

    @property
    def list(self) -> ListSchema:
        return ListSchema()

    @property
    def dict(self) -> DictSchema:
        return DictSchema()

    @property
    def any(self) -> AnySchema:
        return AnySchema()

    @property
    def const(self) -> ConstSchema:
        warnings.warn("Deprecated: 'schema.const' may be removed in the future", FutureWarning)
        return ConstSchema()

    @property
    def bytes(self) -> BytesSchema:
        return BytesSchema()

    @property
    def uuid4(self) -> UUID4Schema:
        return UUID4Schema()

    @property
    def datetime(self) -> DateTimeSchema:
        return DateTimeSchema()

    def __getattr__(self, name: Any) -> Any:
        raise AttributeError(f"{self.__class__.__name__!r} object has no attribute {name!r}")



from datetime import datetime
from uuid import uuid4

from baby_steps import given, then, when
from pytest import raises

from district42 import from_native, schema


def test_none_value():
    with when:
        res = from_native(None)

    with then:
        assert res == schema.none


def test_bool_value():
    with given:
        value = True

    with when:
        res = from_native(value)

    with then:
        assert res == schema.bool(value)


def test_int_value():
    with given:
        value = 42

    with when:
        res = from_native(value)

    with then:
        assert res == schema.int(value)


def test_float_value():
    with given:
        value = 3.14

    with when:
        res = from_native(value)

    with then:
        assert res == schema.float(value)


def test_str_value():
    with given:
        value = "banana"

    with when:
        res = from_native(value)

    with then:
        assert res == schema.str(value)


def test_list_empty_value():
    with given:
        value = []

    with when:
        res = from_native(value)

    with then:
        assert res == schema.list(value)


def test_list_value():
    with given:
        value = [1, 2, 3]

    with when:
        res = from_native(value)

    with then:
        assert res == schema.list([schema.int(1), schema.int(2), schema.int(3)])


def test_dict_empty_value():
    with given:
        value = {}

    with when:
        res = from_native(value)

    with then:
        assert res == schema.dict(value)


def test_dict_value():
    with given:
        value = {
            "id": 1,
            "name": "Bob"
        }

    with when:
        res = from_native(value)

    with then:
        assert res == schema.dict({
            "id": schema.int(1),
            "name": schema.str("Bob"),
        })


def test_bytes_value():
    with given:
        value = b""

    with when:
        res = from_native(value)

    with then:
        assert res == schema.bytes(value)


def test_uuid4_value():
    with given:
        value = uuid4()

    with when:
        res = from_native(value)

    with then:
        assert res == schema.uuid4(value)


def test_datetime_value():
    with given:
        value = datetime.now()

    with when:
        res = from_native(value)

    with then:
        assert res == schema.datetime(value)


def test_unknown_value():
    with given:
        value = Exception()

    with when, raises(Exception) as exception:
        from_native(value)

    with then:
        assert exception.type is ValueError



from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when
from pytest import raises

from district42 import Schema, schema


def test_non_implemented_invert():
    with when, raises(Exception) as exception:
        schema.str.__invert__()

    with then:
        assert exception.type is AttributeError


def test_non_implemented_mod():
    with when, raises(Exception) as exception:
        schema.str.__mod__(None)

    with then:
        assert exception.type is AttributeError


@pytest.mark.parametrize("method", [
    "__invert__",
    "__mod__",
    "__add__",
])
def test_override_invert(method):
    with given:
        mock_ = Mock()
        Schema.__override__(method, mock_)

    with when:
        getattr(schema.str, method)()

    with then:
        assert mock_.mock_calls == [call()]



from datetime import datetime
from typing import Any
from uuid import UUID

from .types import (
    BoolSchema,
    BytesSchema,
    DateTimeSchema,
    DictSchema,
    FloatSchema,
    GenericSchema,
    IntSchema,
    ListSchema,
    NoneSchema,
    StrSchema,
    UUID4Schema,
)

__all__ = ("from_native",)


def from_native(value: Any) -> GenericSchema:
    if value is None:
        return NoneSchema()
    elif isinstance(value, bool):
        return BoolSchema()(value)
    elif isinstance(value, int):
        return IntSchema()(value)
    elif isinstance(value, float):
        return FloatSchema()(value)
    elif isinstance(value, str):
        return StrSchema()(value)
    elif isinstance(value, list):
        return ListSchema()([from_native(x) for x in value])
    elif isinstance(value, dict):
        return DictSchema()({key: from_native(val) for key, val in value.items()})
    elif isinstance(value, bytes):
        return BytesSchema()(value)
    elif isinstance(value, UUID) and (value.version == 4):
        return UUID4Schema()(value)
    elif isinstance(value, datetime):
        return DateTimeSchema()(value)
    else:
        raise ValueError(value)



from baby_steps import given, then, when
from pytest import raises

from district42 import Props, represent
from district42.representor import Representor
from district42.types import Schema


def test_register_visitor():
    with given:
        representation = "schema.custom_type"

        class CustomType(Schema[Props]):
            def __accept__(self, visitor, **kwargs):
                return visitor.visit_custom_type(self, **kwargs)

        class CustomRepresentor(Representor, extend=True):
            def visit_custom_type(self, schema, **kwargs):
                return representation

    with when:
        res = represent(CustomType())

    with then:
        assert res == representation


def test_register_visitor_without_extend():
    with given:
        class AnotherCustomType(Schema[Props]):
            def __accept__(self, visitor, **kwargs):
                return visitor.visit_another_custom_type(self, **kwargs)

        class CustomRepresentor(Representor):
            def visit_another_custom_type(self, schema, **kwargs):
                return "schema.another_custom_type"

    with when, raises(Exception) as exception:
        represent(AnotherCustomType())

    with then:
        assert exception.type is AttributeError



from typing import Any, List, Union

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import (
    DeclarationError,
    make_already_declared_error,
    make_incorrect_len_error,
    make_incorrect_max_len_error,
    make_incorrect_min_len_error,
    make_invalid_type_error,
)
from ..utils import TypeOrEllipsis, is_ellipsis
from ._schema import GenericSchema, Schema

__all__ = ("ListSchema", "ListProps",)


ElementType = TypeOrEllipsis[GenericSchema]


class ListProps(Props):
    @property
    def elements(self) -> Nilable[List[GenericSchema]]:
        return self.get("elements")

    @property
    def type(self) -> Nilable[GenericSchema]:
        return self.get("type")

    @property
    def len(self) -> Nilable[int]:
        return self.get("len")

    @property
    def min_len(self) -> Nilable[int]:
        return self.get("min_len")

    @property
    def max_len(self) -> Nilable[int]:
        return self.get("max_len")


class ListSchema(Schema[ListProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_list(self, **kwargs)

    def __call__(self, /,
                 elements_or_type: Union[List[ElementType], GenericSchema]) -> "ListSchema":
        if not isinstance(elements_or_type, (list, Schema)):
            raise make_invalid_type_error(self, elements_or_type, (list, Schema))

        if (self.props.elements is not Nil) or (self.props.type is not Nil):
            raise make_already_declared_error(self)

        if self.props.len is not Nil:
            raise make_already_declared_error(self)
        if (self.props.min_len is not Nil) or (self.props.max_len is not Nil):
            raise make_already_declared_error(self)

        if isinstance(elements_or_type, Schema):
            return self.__class__(self.props.update(type=elements_or_type))

        for index, element in enumerate(elements_or_type):
            if not isinstance(element, (Schema, type(...))):
                raise make_invalid_type_error(self, element, (Schema, type(...)))
            if is_ellipsis(element):
                if (index != 0) and (index != len(elements_or_type) - 1):
                    raise DeclarationError("`...` must be first or last element")

        if len(elements_or_type) == 2 and \
           is_ellipsis(elements_or_type[0]) and is_ellipsis(elements_or_type[-1]):
            raise DeclarationError("`...` must be first or last element")

        return self.__class__(self.props.update(elements=elements_or_type))

    def __declare_len(self, props: ListProps, length: Any) -> ListProps:
        if not isinstance(length, int):
            raise make_invalid_type_error(self, length, (int,))

        if props.elements is not Nil:
            concrete_elements = [x for x in props.elements if not is_ellipsis(x)]
            if len(props.elements) == len(concrete_elements):
                if length != len(concrete_elements):
                    raise make_incorrect_len_error(self, concrete_elements, length)
            else:
                if length < len(concrete_elements):
                    raise make_incorrect_min_len_error(self, concrete_elements, length)

        return props.update(len=length)

    def __declare_min_len(self, props: ListProps, min_length: Any) -> ListProps:
        if not isinstance(min_length, int):
            raise make_invalid_type_error(self, min_length, (int,))

        if props.elements is not Nil:
            concrete_elements = [x for x in props.elements if not is_ellipsis(x)]
            if min_length > len(concrete_elements):
                raise make_incorrect_min_len_error(self, concrete_elements, min_length)

        return props.update(min_len=min_length)

    def __declare_max_len(self, props: ListProps, max_length: Any) -> ListProps:
        if not isinstance(max_length, int):
            raise make_invalid_type_error(self, max_length, (int,))

        if props.elements is not Nil:
            concrete_elements = [x for x in props.elements if not is_ellipsis(x)]
            if max_length < len(concrete_elements):
                raise make_incorrect_max_len_error(self, concrete_elements, max_length)

        return props.update(max_len=max_length)

    def len(self, /, val_or_min: TypeOrEllipsis[int],
            max: Nilable[TypeOrEllipsis[int]] = Nil) -> "ListSchema":
        if self.props.len is not Nil:
            raise make_already_declared_error(self)

        if (self.props.min_len is not Nil) or (self.props.max_len is not Nil):
            raise make_already_declared_error(self)

        props = self.props
        if is_ellipsis(val_or_min):
            props = self.__declare_max_len(props, max)
        else:
            if max is Nil:
                props = self.__declare_len(props, val_or_min)
            elif is_ellipsis(max):
                props = self.__declare_min_len(props, val_or_min)
            else:
                props = self.__declare_max_len(self.__declare_min_len(props, val_or_min), max)
        return self.__class__(props)



from datetime import datetime
from typing import Any

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import make_already_declared_error, make_invalid_type_error
from ._schema import Schema

__all__ = ("DateTimeSchema", "DateTimeProps",)


class DateTimeProps(Props):
    @property
    def value(self) -> Nilable[datetime]:
        return self.get("value")


class DateTimeSchema(Schema[DateTimeProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_datetime(self, **kwargs)

    def __call__(self, /, value: datetime) -> "DateTimeSchema":
        if not isinstance(value, datetime):
            raise make_invalid_type_error(self, value, (datetime,))

        if self.props.value is not Nil:
            raise make_already_declared_error(self)

        return self.__class__(self.props.update(value=value))



from typing import Any, TypeVar, cast

from niltype import Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ._any_schema import AnySchema
from ._schema import GenericSchema, Schema

__all__ = ("TypeAliasSchema", "TypeAliasProps",
           "GenericTypeAliasSchema", "TypeAliasPropsType",)


class TypeAliasProps(Props):
    @property
    def type(self) -> GenericSchema:
        return cast(GenericSchema, self.get("type", AnySchema()))

    @property
    def name(self) -> Nilable[str]:
        return self.get("name")


TypeAliasPropsType = TypeVar("TypeAliasPropsType", bound=TypeAliasProps)


class GenericTypeAliasSchema(Schema[TypeAliasPropsType]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_type_alias(self, **kwargs)


class TypeAliasSchema(GenericTypeAliasSchema[TypeAliasProps]):
    pass



from typing import Any, Iterator, Mapping, TypeVar

from niltype import Nil, Nilable

__all__ = ("Props", "PropsType",)

PropsType = TypeVar("PropsType", bound="Props")


class Props:
    def __init__(self, registry: Nilable[Mapping[str, Any]] = Nil) -> None:
        self._registry = registry if (registry is not Nil) else {}

    def get(self, name: str, default: Nilable[Any] = Nil) -> Nilable[Any]:
        return self._registry.get(name, default)

    def set(self: PropsType, name: str, value: Any) -> PropsType:
        registry = {**self._registry, name: value}
        return self.__class__(registry)

    def update(self: PropsType, **keys: Any) -> PropsType:
        registry = {**self._registry, **keys}
        return self.__class__(registry)

    def __iter__(self) -> Iterator[str]:
        return iter(self._registry)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._registry}>"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        for key, val in self._registry.items():
            other_val = other.get(key)
            if val != other_val:
                return False

        for key, other_val in other._registry.items():
            val = self.get(key)
            if other_val != val:
                return False

        return True



from typing import Any, Dict, Union

from ..types._optional import optional
from ._is_ellipsis import is_ellipsis

__all__ = ("rollout",)


KeysType = Dict[Union[str, optional], Any]


def rollout(keys: KeysType, *, separator: str = ".") -> KeysType:
    updated: KeysType = {}

    for comp_key, val in keys.items():
        if is_ellipsis(comp_key):
            if not is_ellipsis(val):
                raise ValueError("Expected both key and value to be ellipsis")
            updated[comp_key] = val
            continue

        is_optional = False
        if isinstance(comp_key, optional):
            comp_key = comp_key.key
            is_optional = True
        if not isinstance(comp_key, str):
            raise TypeError(f"Unexpected key type {type(comp_key)}")

        parts = comp_key.split(separator)
        key = parts[0]
        if len(parts) == 1:
            updated[optional(key) if is_optional else key] = val
        else:
            if key not in updated:
                updated[key] = {}
            tail = separator.join(parts[1:])
            updated[key][optional(tail) if is_optional else tail] = val

    for k, v in updated.items():
        updated[k] = rollout(v, separator=separator) if isinstance(v, dict) else v

    return updated



from typing import Any
from uuid import UUID

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import make_already_declared_error, make_invalid_type_error
from ._schema import Schema

__all__ = ("UUID4Schema", "UUID4Props",)


class UUID4Props(Props):
    @property
    def value(self) -> Nilable[UUID]:
        return self.get("value")


class UUID4Schema(Schema[UUID4Props]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_uuid4(self, **kwargs)

    def __call__(self, /, value: UUID) -> "UUID4Schema":
        if not isinstance(value, UUID):
            raise make_invalid_type_error(self, value, (UUID,))

        if self.props.value is not Nil:
            raise make_already_declared_error(self)

        return self.__class__(self.props.update(value=value))



import re
from typing import Any

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import (
    DeclarationError,
    make_already_declared_error,
    make_incorrect_len_error,
    make_incorrect_max_len_error,
    make_incorrect_min_len_error,
    make_invalid_type_error,
)
from ..utils import TypeOrEllipsis, is_ellipsis
from ._schema import Schema

__all__ = ("StrSchema", "StrProps",)


class StrProps(Props):
    @property
    def value(self) -> Nilable[str]:
        return self.get("value")

    @property
    def len(self) -> Nilable[int]:
        return self.get("len")

    @property
    def min_len(self) -> Nilable[int]:
        return self.get("min_len")

    @property
    def max_len(self) -> Nilable[int]:
        return self.get("max_len")

    @property
    def alphabet(self) -> Nilable[str]:
        return self.get("alphabet")

    @property
    def substr(self) -> Nilable[str]:
        return self.get("substr")

    @property
    def pattern(self) -> Nilable[str]:
        return self.get("pattern")


class StrSchema(Schema[StrProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_str(self, **kwargs)

    def __call__(self, /, value: str) -> "StrSchema":
        if not isinstance(value, str):
            raise make_invalid_type_error(self, value, (str,))

        if self.props.value is not Nil:
            raise make_already_declared_error(self)

        if self.props.len is not Nil:
            raise make_already_declared_error(self)

        if (self.props.min_len is not Nil) or (self.props.max_len is not Nil):
            raise make_already_declared_error(self)

        if self.props.alphabet is not Nil:
            raise make_already_declared_error(self)

        if self.props.substr is not Nil:
            raise make_already_declared_error(self)

        if self.props.pattern is not Nil:
            raise make_already_declared_error(self)

        return self.__class__(self.props.update(value=value))

    def __declare_len(self, props: StrProps, length: Any) -> StrProps:
        if not isinstance(length, int):
            raise make_invalid_type_error(self, length, (int,))

        if (props.value is not Nil) and (len(props.value) != length):
            raise make_incorrect_len_error(self, props.value, length)

        return props.update(len=length)

    def __declare_min_len(self, props: StrProps, min_length: Any) -> StrProps:
        if not isinstance(min_length, int):
            raise make_invalid_type_error(self, min_length, (int,))

        if (props.value is not Nil) and (min_length > len(props.value)):
            raise make_incorrect_min_len_error(self, props.value, min_length)

        return props.update(min_len=min_length)

    def __declare_max_len(self, props: StrProps, max_length: Any) -> StrProps:
        if not isinstance(max_length, int):
            raise make_invalid_type_error(self, max_length, (int,))

        if (props.value is not Nil) and (max_length < len(props.value)):
            raise make_incorrect_max_len_error(self, props.value, max_length)

        return props.update(max_len=max_length)

    def len(self, /, val_or_min: TypeOrEllipsis[int],
            max: Nilable[TypeOrEllipsis[int]] = Nil) -> "StrSchema":
        if self.props.len is not Nil:
            raise make_already_declared_error(self)

        if (self.props.min_len is not Nil) or (self.props.max_len is not Nil):
            raise make_already_declared_error(self)

        if self.props.pattern is not Nil:
            raise make_already_declared_error(self)

        props = self.props
        if is_ellipsis(val_or_min):
            props = self.__declare_max_len(props, max)
        else:
            if max is Nil:
                props = self.__declare_len(props, val_or_min)
            elif is_ellipsis(max):
                props = self.__declare_min_len(props, val_or_min)
            else:
                props = self.__declare_max_len(self.__declare_min_len(props, val_or_min), max)
        return self.__class__(props)

    def alphabet(self, /, letters: str) -> "StrSchema":
        if not isinstance(letters, str):
            raise make_invalid_type_error(self, letters, (str,))

        if self.props.alphabet is not Nil:
            raise make_already_declared_error(self)

        if self.props.pattern is not Nil:
            raise make_already_declared_error(self)

        if self.props.value is not Nil:
            missing_letters = {x for x in self.props.value if x not in letters}
            if len(missing_letters) > 0:
                message = f"`{self!r}` alphabet is missing letters: "
                message += repr("".join(sorted(missing_letters)))
                raise DeclarationError(message)

        return self.__class__(self.props.update(alphabet=letters))

    def contains(self, substr: str) -> "StrSchema":
        if not isinstance(substr, str):
            raise make_invalid_type_error(self, substr, (str,))

        if self.props.substr is not Nil:
            raise make_already_declared_error(self)

        if self.props.pattern is not Nil:
            raise make_already_declared_error(self)

        if self.props.value is not Nil:
            if substr not in self.props.value:
                message = f"`{self!r}` does not contain {substr!r}"
                raise DeclarationError(message)

        return self.__class__(self.props.update(substr=substr))

    def regex(self, pattern: str) -> "StrSchema":
        if not isinstance(pattern, str):
            raise make_invalid_type_error(self, pattern, (str,))

        if (self.props.pattern is not Nil) or (self.props.alphabet is not Nil) or \
           (self.props.len is not Nil) or (self.props.min_len is not Nil) or \
           (self.props.len is not Nil) or (self.props.substr is not Nil):
            raise make_already_declared_error(self)

        try:
            re.compile(pattern)
        except re.error as e:
            message = f"Invalid pattern ({e})"
            raise DeclarationError(message) from None

        if self.props.value is not Nil:
            if re.search(pattern, self.props.value) is None:
                message = f"`{self!r}` does not match {pattern!r}"
                raise DeclarationError(message)

        return self.__class__(self.props.update(pattern=pattern))



from typing import Any, Type, TypeVar, cast

from ._from_native import from_native
from ._props import Props
from ._schema_facade import SchemaFacade
from ._schema_visitor import SchemaVisitor, SchemaVisitorReturnType
from ._version import version
from .representor import Representor
from .types import AnySchema, GenericSchema, Schema, make_required, optional

__version__ = version
__all__ = ("schema", "GenericSchema", "Props", "SchemaVisitor", "SchemaVisitorReturnType",
           "from_native", "optional", "register_type", "represent", "make_required", )


schema = SchemaFacade()
_representor = Representor()

_SchemaType = TypeVar("_SchemaType", bound=GenericSchema)


def register_type(name: str, schema_type: Type[_SchemaType]) -> _SchemaType:
    if not issubclass(schema_type, Schema):
        raise TypeError(f"{schema_type!r} must be a subclass of Schema")
    setattr(SchemaFacade, name, property(lambda self: schema_type()))
    return cast(_SchemaType, getattr(schema, name))


def represent(self: GenericSchema, **kwargs: Any) -> str:
    return self.__accept__(_representor, **kwargs)


def union(self: GenericSchema, other: Any) -> AnySchema:
    return schema.any(self, other)


Schema.__override__("__repr__", represent)
Schema.__override__("__or__", union)



from typing import Any, Dict, Generator, KeysView, List, Optional, Set, Tuple, Union

from niltype import Nil, Nilable

from .._props import Props
from .._schema_visitor import SchemaVisitor
from .._schema_visitor import SchemaVisitorReturnType as ReturnType
from ..errors import DeclarationError, make_already_declared_error, make_invalid_type_error
from ..utils import TypeOrEllipsis, is_ellipsis
from ._optional import optional
from ._schema import GenericSchema, Schema

__all__ = ("DictSchema", "DictProps", "make_required", "optional",)


class DictProps(Props):
    @property
    def keys(self) -> Nilable[Dict[Any, Tuple[GenericSchema, bool]]]:
        return self.get("keys")


class DictSchema(Schema[DictProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return visitor.visit_dict(self, **kwargs)

    def __call__(self, /, keys: Dict[Any, TypeOrEllipsis[GenericSchema]]) -> "DictSchema":
        if not isinstance(keys, dict):
            raise make_invalid_type_error(self, keys, (dict,))

        if self.props.keys is not Nil:
            raise make_already_declared_error(self)

        real_keys = {}
        for key, val in keys.items():
            if is_ellipsis(key) or is_ellipsis(val):
                if not is_ellipsis(key):
                    message = f"Inappropriate type of key {key!r} ({type(key)!r})"
                    raise DeclarationError(message)
                if not is_ellipsis(val):
                    message = f"Inappropriate type of value {val!r} ({type(val)!r})"
                    raise DeclarationError(message)
            else:
                if not isinstance(val, Schema):
                    raise make_invalid_type_error(self, val, (Schema,))
            if isinstance(key, optional):
                real_keys[key.key] = (val, True)
            else:
                real_keys[key] = (val, False)

        return self.__class__(self.props.update(keys=real_keys))

    def __getitem__(self, /, key: Any) -> GenericSchema:
        if (self.props.keys is Nil) or (key not in self.props.keys) or (is_ellipsis(key)):
            key_repr = "..." if is_ellipsis(key) else key
            raise KeyError(key_repr)
        return self.props.keys[key][0]

    def __add__(self, /, other: "DictSchema") -> "DictSchema":
        if not isinstance(other, DictSchema):
            raise TypeError(
                f"Unsupported operand type for +: '{self.__class__.__name__}' and {type(other)!r}")

        self_keys = self.props.keys if (self.props.keys is not Nil) else {}
        other_keys = other.props.keys if (other.props.keys is not Nil) else {}
        merged_keys = {**self_keys, **other_keys}
        return self.__class__(self.props.update(keys=merged_keys))

    def keys(self) -> KeysView[Any]:
        if self.props.keys is Nil:
            return {}.keys()
        return self.props.keys.keys()

    def __iter__(self) -> Generator[Any, None, None]:
        yield from self.keys()


RequiredKeysType = Union[Set[str], List[str], Tuple[str, ...]]


def make_required(schema: DictSchema, keys: Optional[RequiredKeysType] = None) -> DictSchema:
    if not isinstance(schema, DictSchema):
        message = f"Inappropriate type of schema {schema!r} ({type(schema)!r})"
        raise DeclarationError(message)

    if not isinstance(keys, (set, list, tuple, type(None))):
        message = f"Inappropriate type of keys {keys!r} ({type(keys)!r})"
        raise DeclarationError(message)

    if keys is None:
        keys = set(schema.keys())

    props_keys = schema.props.keys if (schema.props.keys is not Nil) else {}
    for key in keys:
        if key not in props_keys:
            message = f"Nonexisting key {key!r}"
            raise DeclarationError(message)

    if schema.props.keys is Nil:
        return schema
    else:
        updated_keys = {}
        for key, (val, is_optional) in props_keys.items():
            updated_keys[key] = (val, False if (key in keys) else is_optional)
        return schema.__class__(schema.props.update(keys=updated_keys))



from baby_steps import then, when

from district42 import schema


def test_schema_eq():
    with when:
        res = (schema.int == schema.int)

    with then:
        assert res is True


def test_schema_not_eq():
    with when:
        res = (schema.int == schema.str)

    with then:
        assert res is False


def test_schema_ne():
    with when:
        res = (schema.int != schema.str)

    with then:
        assert res is True


def test_schema_not_ne():
    with when:
        res = (schema.int != schema.int)

    with then:
        assert res is False



from typing import Any

from ..utils import is_ellipsis

__all__ = ("optional",)


class optional:
    def __init__(self, key: Any) -> None:
        if is_ellipsis(key):
            raise TypeError(key)
        self._key = key

    @property
    def key(self) -> Any:
        return self._key

    def __repr__(self) -> str:
        return f"optional({self._key!r})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self._key == other.key)

    def __hash__(self) -> int:
        return hash((self._key,))



from ._any_schema import AnyProps, AnySchema
from ._bool_schema import BoolProps, BoolSchema
from ._bytes_schema import BytesProps, BytesSchema
from ._const_schema import ConstProps, ConstSchema
from ._datetime_schema import DateTimeProps, DateTimeSchema
from ._dict_schema import DictProps, DictSchema, make_required, optional
from ._float_schema import FloatProps, FloatSchema
from ._int_schema import IntProps, IntSchema
from ._list_schema import ListProps, ListSchema
from ._none_schema import NoneProps, NoneSchema
from ._schema import GenericSchema, Schema
from ._str_schema import StrProps, StrSchema
from ._type_alias_schema import (
    GenericTypeAliasSchema,
    TypeAliasProps,
    TypeAliasPropsType,
    TypeAliasSchema,
)
from ._uuid4_schema import UUID4Props, UUID4Schema

__all__ = ("AnyProps", "AnySchema", "BoolProps", "BoolSchema", "BytesProps", "BytesSchema",
           "ConstProps", "ConstSchema", "DictProps", "DictSchema", "FloatProps", "FloatSchema",
           "IntProps", "IntSchema", "ListProps", "ListSchema", "NoneProps", "NoneSchema",
           "StrProps", "StrSchema", "UUID4Props", "UUID4Schema", "DateTimeProps", "DateTimeSchema",
           "TypeAliasSchema", "TypeAliasProps",
           "GenericTypeAliasSchema", "TypeAliasPropsType", "GenericSchema", "Schema",
           "optional", "make_required", )



from baby_steps import given, then, when
from pytest import raises

from district42 import Props, register_type, schema
from district42.types import Schema


def test_register_type():
    with given:
        class CustomType(Schema[Props]):
            def __accept__(self, visitor, **kwargs):
                pass

    with when:
        res = register_type("custom_type", CustomType)

    with then:
        assert isinstance(res, CustomType)
        assert isinstance(schema.custom_type, CustomType)
        assert schema.custom_type is not schema.custom_type


def test_register_incorrect_type():
    with given:
        class CustomIncorrectType:
            def __accept__(self, visitor, **kwargs):
                pass

    with when, raises(Exception) as exception:
        register_type("custom_incorrect_type", CustomIncorrectType)

    with then:
        assert exception.type is TypeError
        assert str(exception.value) == f"{CustomIncorrectType!r} must be a subclass of Schema"
